import typing as typ
from .imgui_render_window import RendererBackend, register_backend
from vtkmodules.vtkCommonCore import vtkCommand
from vtkmodules.vtkRenderingCore import vtkRenderWindow, vtkRenderWindowInteractor
import typing as typ
try:
    import imgui
except ImportError:
    imgui = None

__all__ = ['RendererBackendPyImgui']

@register_backend("pyimgui")
class RendererBackendPyImgui(RendererBackend):
    def __init__(self, 
                 interactor: vtkRenderWindowInteractor, 
                 render_window: vtkRenderWindow,
                 border = False) -> None:
        if not imgui:
            raise ModuleNotFoundError(f"{self.__class__.__name__} requires the 'pyimgui' package.")
        super().__init__(interactor, render_window, border=border)

    def render(self):
        #if size is None:
        # get the maximum available size
        size = imgui.get_content_region_available()
        size = (size.x, size.y)
        self.render_window.size = size
        self.render_window.render()

        # adjust the size of this interactor as well
        self.interactor.SetSize(int(size[0]), int(size[1]))
        imgui.push_style_var(imgui.STYLE_WINDOW_PADDING, (0, 0))
        no_scroll_flags = imgui.WINDOW_NO_SCROLLBAR | imgui.WINDOW_NO_SCROLL_WITH_MOUSE
        imgui.begin_child("##Viewport", size[0], size[1], self.border, no_scroll_flags)
        image_size = imgui.get_content_region_available()
        imgui.image(self.render_window.texture_id, 
                    image_size.x, image_size.y,
                    (0, 1), (1, 0))
        # process the events of this widget
        self.process_events()
        imgui.end_child()
        imgui.pop_style_var()

    def _process_mouse_events(self, io):
        if imgui.is_window_hovered():
            if imgui.is_mouse_clicked(imgui.MOUSE_BUTTON_LEFT):
                self.interactor.InvokeEvent(vtkCommand.LeftButtonPressEvent)
            elif imgui.is_mouse_clicked(imgui.MOUSE_BUTTON_RIGHT):
                self.interactor.InvokeEvent(vtkCommand.RightButtonPressEvent)
            elif imgui.is_mouse_clicked(imgui.MOUSE_BUTTON_MIDDLE):
                self.interactor.InvokeEvent(vtkCommand.MiddleButtonPressEvent)
            elif io.mouse_wheel > 0:
                self.interactor.InvokeEvent(vtkCommand.MouseWheelForwardEvent)
            elif io.mouse_wheel < 0:
                self.interactor.InvokeEvent(vtkCommand.MouseWheelBackwardEvent)

        if imgui.is_mouse_released(imgui.MOUSE_BUTTON_LEFT):
            self.interactor.InvokeEvent(vtkCommand.LeftButtonReleaseEvent)
        elif imgui.is_mouse_released(imgui.MOUSE_BUTTON_RIGHT):
            self.interactor.InvokeEvent(vtkCommand.RightButtonReleaseEvent)
        elif imgui.is_mouse_released(imgui.MOUSE_BUTTON_MIDDLE):
            self.interactor.InvokeEvent(vtkCommand.MiddleButtonReleaseEvent)
        
        self.interactor.InvokeEvent(vtkCommand.MouseMoveEvent)

    def process_events(self):
        """
        Handle events by passing them to the underlying vtk interactor. This method is called automatically
        on rendering.
        """
        # do nothing as long as the mouse pointer is not within the current window or it is not focussed
        if not imgui.is_window_focused() and not imgui.is_window_hovered():
            return
        io = imgui.get_io()
        io.config_windows_move_from_title_bar_only = True # do not drag the window when clicking on the image
        viewport_pos = imgui.get_cursor_start_pos()

        xpos = int(io.mouse_pos.x - viewport_pos.x)
        ypos = int(io.mouse_pos.y - viewport_pos.y)

        ctrl = io.key_ctrl
        shift = io.key_shift
        
        repeat = 0
        #if io.mouse_double_clicked[0] or io.mouse_double_clicked[1] or io.mouse_double_clicked[2]:
        #    repeat = 1

        if xpos < 0 or ypos < 0:
            return 
        
        self.interactor.SetEventInformationFlipY(xpos, ypos, ctrl, shift, chr(0), repeat, None)

        self._process_mouse_events(io)
        # no keyboard events yet
        #self._process_keyboard_events(xpos, ypos, ctrl, shift)

    def show(self,
             title: typ.Optional[str] = None, 
             window_size: tuple[int, int] = (1400, 1080)):
        """
        Show method for the pyimgui package

        Parameters
        ----------
        title, optional
            The title of the window, if None, a default title is used.
        window_size, optional
            The size of the displayed window, by default (1400, 1080)
        """
        import glfw
        import OpenGL.GL as GL
        import imgui
        from imgui.integrations.glfw import GlfwRenderer

        if not glfw.init():
            raise ValueError("Could not initialize OpenGL context")

        # Create a window and its OpenGL context
        window = glfw.create_window(int(window_size[0]), int(window_size[1]), title or "ImguiPlotter", None, None)
        glfw.make_context_current(window)

        if not window:
            glfw.terminate()
            raise ValueError("Could not initialize Window")
        
        background_color = (0.0, 0.0, 0.0, 1.0)
        GL.glClearColor(*background_color)
        imgui.create_context()

        impl = GlfwRenderer(window)

        while not glfw.window_should_close(window):
            glfw.poll_events()
            impl.process_inputs()
            imgui.new_frame()
            vec = imgui.get_main_viewport().pos
            imgui.set_next_window_position(vec.x, vec.y, imgui.ONCE)
            size = imgui.get_main_viewport().size
            imgui.set_next_window_size(size.x, size.y)
            imgui.set_next_window_bg_alpha(1.0)
            window_flags = imgui.WINDOW_NO_BRING_TO_FRONT_ON_FOCUS | \
                           imgui.WINDOW_NO_TITLE_BAR | \
                           imgui.WINDOW_NO_DECORATION | \
                           imgui.WINDOW_NO_RESIZE | \
                           imgui.WINDOW_NO_MOVE
                                 
            imgui.begin("Vtk Viewer", flags=window_flags)
            self.render()
            imgui.end()

            imgui.render()

            GL.glClearColor(*background_color)
            GL.glClear(GL.GL_COLOR_BUFFER_BIT)
            impl.render(imgui.get_draw_data())
            glfw.swap_buffers(window)

        impl.shutdown()
        glfw.terminate()
      


from __future__ import annotations
from vtkmodules.util import numpy_support as VN
from vtkmodules.vtkCommonCore import vtkFloatArray
from vtkmodules.vtkFiltersSources import vtkSphereSource
from vtkmodules.vtkFiltersCore import vtkTriangleMeshPointNormals
from vtkmodules.vtkRenderingOpenGL2 import vtkOpenGLPolyDataMapper
from vtkmodules.vtkCommonTransforms import vtkTransform

import numpy as np
from numpy.typing import ArrayLike
from pyvista.plotting.render_window_interactor import _style_factory
import pyvista as pv
from typing import Any, Sequence


class DefaultInteractorKeybindsMixin:
    def __init__(self) -> None:
        self.add_observer('CharEvent', self.on_char_event)
        self.__key_dispatcher = {
            'f': self._fly_to_point,
            'r': self._reset_camera,
            'p': self._pick,
        }

    def on_char_event(self, obj: Any, event: Any) -> None:
        interactor = self.GetInteractor()
        key = interactor.GetKeyCode()
        event_pos = interactor.GetEventPosition()
        self.FindPokedRenderer(event_pos[0], event_pos[1])

        try:
            self.__key_dispatcher[key]()
        except KeyError:
            pass

    def _fly_to_point(self) -> None:
        interactor = self.GetInteractor()
        event_pos = interactor.GetEventPosition()
        self._set_pivot(event_pos)
        self._show_pivot_sphere()

        interactor.SetDolly(1.0)
        interactor.FlyTo(self.GetCurrentRenderer(), self.pivot)
        self._hide_pivot_sphere()

    def _reset_camera(self) -> None:
        renderer = self.GetCurrentRenderer()
        if renderer:
            renderer.ResetCamera()
        self.GetInteractor().Render()

    def _pick(self) -> None:
        interactor = self.GetInteractor()
        renderer = self.GetCurrentRenderer()
        if renderer:
            interactor.StartPickCallback()
            picker = interactor.GetPicker()
            if picker:
                event_pos = interactor.GetEventPosition()
                picker.Pick(event_pos[0], event_pos[1], 0.0, renderer)
            interactor.EndPickCallback()


__all__ = ["InteractorStylePivot"]

def _normalize(v: ArrayLike) -> ArrayLike:
    """Normalize the array

    Args:
        v: _description_

    Returns:
        _description_
    """
    v = np.asarray(v)
    norm = np.linalg.norm(v)
    if norm == 0:
        return v
    return v / norm


def _create_pivot_sphere(radius: float = 0.25, resolution: int = 16) -> pv.Actor:
    sphere = vtkSphereSource()
    sphere.SetRadius(radius)
    sphere.SetThetaResolution(resolution)
    sphere.SetPhiResolution(resolution)
    norms = vtkTriangleMeshPointNormals()
    norms.SetInputConnection(sphere.GetOutputPort())
    norms.Update()
    sphere_mapper = vtkOpenGLPolyDataMapper()
    sphere_mapper.SetInputConnection(norms.GetOutputPort())
    del sphere

    pivot_sphere = pv.Actor()
    pivot_sphere.mapper = sphere_mapper

    pivot_sphere.GetProperty().SetColor(1, 1, 1)
    pivot_sphere.prop.SetRepresentationToSurface()
    pivot_sphere.prop.SetAmbientColor(0.2, 0.2, 1.0)
    pivot_sphere.prop.SetDiffuseColor(1.0, 0.65, 0.7)
    pivot_sphere.prop.SetSpecularColor(1.0, 1.0, 1.0)
    pivot_sphere.prop.SetSpecular(0.5)
    pivot_sphere.prop.SetDiffuse(0.7)
    pivot_sphere.prop.SetAmbient(0.5)
    pivot_sphere.prop.SetSpecularPower(20.0)
    pivot_sphere.prop.SetOpacity(1.0)
    #sp = pivot_sphere.GetShaderProperty()

    # modify the shader to generate the normal colors
    # sp.AddVertexShaderReplacement(
    #     "//VTK::Normal::Dec",
    #     True,
    #     "//VTK::Normal::Dec\n"
    #     "  varying vec3 myNormalMCVSOutput;\n",
    #     False)
    # sp.AddVertexShaderReplacement(
    #     "//VTK::Normal::Impl",
    #     True,
    #     "//VTK::Normal::Impl\n"
    #     "  myNormalMCVSOutput = normalMC;\n",
    #     False)
    # sp.AddFragmentShaderReplacement(
    #     "//VTK::Normal::Dec",
    #     True,
    #     "//VTK::Normal::Dec\n"
    #     "  varying vec3 myNormalMCVSOutput;\n",
    #     False)
    # sp.AddFragmentShaderReplacement(
    #     "//VTK::Normal::Impl",
    #     True,
    #     "//VTK::Normal::Impl\n"
    #     "  diffuseColor = abs(myNormalMCVSOutput);\n",
    #     False)

    del sphere_mapper

    return pivot_sphere


_TrackBallStyle = _style_factory('TrackballCamera')


class InteractorStylePivot(DefaultInteractorKeybindsMixin, _TrackBallStyle):
    """
    Custom interactor style that does every interaction relative to a pivot point under the mouse pointer.
    The pivot point is picked on every mouse click using a window size dependent tolerance. If there is no point under
    the mouse the previous pivot point is used.

    Args:
        parent: The parent of this instance
        pivot_sphere_radius: The radius of the pivot sphere. Defaults to 0.25.
        pivot_sphere_resolution: The number of segments used to draw the sphere. Defaults to 16.
    """
    def __init__(self, parent: Any, 
                 pivot_sphere_radius: float = 0.25, 
                 pivot_sphere_resolution: int = 16) -> None:
        _TrackBallStyle.__init__(self, parent)
        DefaultInteractorKeybindsMixin.__init__(self)
        self.remove_observers()
        self.add_observer("LeftButtonPressEvent", self._left_button_press_event)
        self.add_observer("LeftButtonReleaseEvent", self._left_button_release_event)
        self.add_observer("RightButtonPressEvent", self._right_button_press_event)
        self.add_observer("RightButtonReleaseEvent", self._right_button_release_event)
        self.add_observer("MouseMoveEvent", self._mouse_move_event)
        self.add_observer("MiddleButtonPressEvent", self._middle_button_press_event)
        self.add_observer("MiddleButtonReleaseEvent", self._middle_button_release_event)
        self.add_observer("MouseWheelForwardEvent", self._mousewheel_forward_event)
        self.add_observer("MouseWheelBackwardEvent", self._mousewheel_backward_event)

        # create focus sphere actor
        self._pivot_sphere_radius = pivot_sphere_radius
        self._pivot_sphere_resolution = pivot_sphere_resolution
        self._pivot_sphere_visible = False

        self._pivot_sphere = _create_pivot_sphere(pivot_sphere_radius, pivot_sphere_resolution)

        self.left_down = False
        self.right_down = False
        self.state = 0
        self.pivot = [0, 0, 0]
        self.tolerance = 0.002

    @property
    def pivot_sphere_radius(self) -> float:
        return self._pivot_sphere_radius

    @pivot_sphere_radius.setter
    def pivot_sphere_radius(self, radius: float) -> None:
        show_again = False
        if self._pivot_sphere_visible:
            show_again = True
            self._hide_pivot_sphere()
        self._pivot_sphere_radius = radius
        self._pivot_sphere = _create_pivot_sphere(self._pivot_sphere_radius, self._pivot_sphere_resolution)
        if show_again:
            self._show_pivot_sphere()

    @property
    def pivot_sphere_resolution(self) -> int:
        return self._pivot_sphere_resolution

    @pivot_sphere_resolution.setter
    def pivot_sphere_resolution(self, resolution: int) -> None:
        show_again = False
        if self._pivot_sphere_visible:
            show_again = True
            self._hide_pivot_sphere()
        self._pivot_sphere_resolution = resolution
        self._pivot_sphere = _create_pivot_sphere(self._pivot_sphere_radius, self._pivot_sphere_resolution)
        if show_again:
            self._show_pivot_sphere()

    @property
    def tolerance(self) -> float:
        """
        Tolerance of the mouse pointer in percentage of the window diagonal size.

        Returns:
            the tolerance parameter

        """
        return self.__tolerance

    @tolerance.setter
    def tolerance(self, tol: float) -> None:
        """
        Tolerance of the mouse pointer in percentage of the window diagonal size.

        Args:
            tol: the new tolerance
        """
        self.__tolerance = tol

    @property
    def translation_plane_normal(self) -> np.ndarray:
        """
        EXPERIMENTAL
        The plane normal used for the translation.

        Returns:
            The normal vector
        """
        return getattr(self, '_translation_plane_normal', None)

    @translation_plane_normal.setter
    def translation_plane_normal(self, plane: ArrayLike) -> None:
        """        
        EXPERIMENTAL
        The plane normal used for the translation

        Args:
            plane: the plane normal
        """
        plane = np.asarray(plane).ravel()
        if len(plane) != 3:
            raise ValueError("Plane must be a normalized vector of 3 components")
        self._translation_plane_normal = plane

    def _set_pivot(self, event_pos: tuple[int, int]) -> None:
        """
        Pick a new pivot point based on the event position.
        The point is determined by extracting an area from the renderer's depth buffer
        and choosing the pixel position closest to the event position. This position is then projected
        into world coordinates.

        Args:
            event_pos: the event position in display coordinates

        Returns:
            True, if a new pivot point has been picked, False otherwise
        """
        self.FindPokedRenderer(event_pos[0], event_pos[1])
        renwin = self.GetInteractor().GetRenderWindow()
        win_size = np.asarray(renwin.GetSize())
        diagonal_len = np.sqrt(np.sum(win_size**2))
        vfa = vtkFloatArray()
        extent = int(np.ceil(self.tolerance * diagonal_len))
        renwin.GetZbufferData(event_pos[0] - extent, event_pos[1] - extent, event_pos[0]+extent, event_pos[1]+extent, vfa)
        size = extent * 2 + 1
        data = VN.vtk_to_numpy(vfa).reshape(size, size) - 1.0

        nonzero = np.nonzero(data)
        if len(nonzero[0]) > 0:
            distances_screen = (nonzero[0] - extent) ** 2 + (nonzero[1] - extent) ** 2
            distances_scene = data[nonzero]

            distances_total = distances_scene * (distances_screen + 1 + 100) #distances_screen * 0.5 + distances_scene * 0.5

            #print("distances_scene", distances_scene.min(), distances_scene.max())
            #print("distances_screen", distances_screen.min(), distances_screen.max())
            min_id = np.argmin(distances_total)
            nearest_index = (nonzero[0][min_id], nonzero[1][min_id])
            event_pos = (event_pos[0] - extent + nearest_index[0], event_pos[1] - extent + nearest_index[1])
            renderer = self.GetCurrentRenderer()
            world = np.zeros(4)
            self.ComputeDisplayToWorld(renderer, *event_pos, data[nearest_index] + 1.0, world)
            for i in range(3):
                self.pivot[i] = world[i] / world[3]
            return True
        return False

    def _scale_pivot_sphere(self) -> None:
        camera = self.GetCurrentRenderer().GetActiveCamera()
        from_pos = camera.GetPosition()
        vec = np.asarray(self.pivot) - np.asarray(from_pos)
        at_v = _normalize(camera.GetDirectionOfProjection())
        s = 0.02 * np.dot(at_v, vec)
        self._pivot_sphere.SetScale(s, s, s)

    def _show_pivot_sphere(self) -> None:
        """
        Show a sphere at the pivot position
        """
        self._pivot_sphere.SetPosition(self.pivot)
        # calculate scale so focus sphere always is the same size on the screen
        self._scale_pivot_sphere()
        self._focus_sphere_renderer = self.GetCurrentRenderer()
        self._focus_sphere_renderer.AddActor(self._pivot_sphere)
        self._pivot_sphere_visible = True

        self.GetInteractor().Render()

    def _hide_pivot_sphere(self) -> None:
        """
        Hide the sphere if it is visible
        """
        sphere_renderer = getattr(self, '_focus_sphere_renderer', None)
        if sphere_renderer:
            sphere_renderer.RemoveActor(self._pivot_sphere)
            self._pivot_sphere_visible = False

    def _left_button_press_event(self, obj: Any, event: Any) -> None:
        click_pos = self.GetInteractor().GetEventPosition()
        self.left_down = True
        if not self.right_down:
            self._set_pivot(click_pos)
            self._show_pivot_sphere()
        self.last_pos = click_pos

        super().OnLeftButtonDown()

    def _right_button_press_event(self, obj: Any, event: Any) -> None:
        click_pos = self.GetInteractor().GetEventPosition()
        self.right_down = True
        if not self.left_down:
            self._set_pivot(click_pos)
            self._show_pivot_sphere()
        self.last_pos = click_pos
        super().OnMiddleButtonDown()

    def _middle_button_press_event(self, obj: Any, event: Any) -> None:
        click_pos = self.GetInteractor().GetEventPosition()
        self._set_pivot(click_pos)
        self._show_pivot_sphere()
        self.last_pos = click_pos
        super().OnRightButtonDown()

    def _left_button_release_event(self, obj: Any, event: Any) -> None:
        self.left_down = False
        if not self.right_down:
            self._hide_pivot_sphere()
        super().OnLeftButtonUp()

    def _right_button_release_event(self, obj: Any, event: Any) -> None:
        self.right_down = False
        if not self.left_down:
            self._hide_pivot_sphere()
        super().OnMiddleButtonUp()

    def _middle_button_release_event(self, obj: Any, event: Any) -> None:
        self._hide_pivot_sphere()
        super().OnRightButtonUp()

    def _mousewheel_forward_event(self, obj: Any, event: Any) -> None:
        event_pos = self.GetInteractor().GetEventPosition()
        self.FindPokedRenderer(event_pos[0], event_pos[1])
        ret = self._set_pivot(event_pos)
        if ret != 0:
            camera = self.GetCurrentRenderer().GetActiveCamera()
            self.StartDolly()
            delta = self.GetMotionFactor() * -0.005 * self.GetMouseWheelMotionFactor()
            if camera.GetParallelProjection():
                camera.SetParallelScale(camera.GetParallelScale / delta)
            else:
                self._dolly_camera(delta)
            self.EndDolly()

    def _mousewheel_backward_event(self, obj: Any, event: Any) -> None:
        event_pos = self.GetInteractor().GetEventPosition()
        self.FindPokedRenderer(event_pos[0], event_pos[1])
        ret = self._set_pivot(event_pos)
        if ret != 0:
            camera = self.GetCurrentRenderer().GetActiveCamera()
            self.StartDolly()
            delta = self.GetMotionFactor() * 0.005 * self.GetMouseWheelMotionFactor()
            if camera.GetParallelProjection():
                camera.SetParallelScale(camera.GetParallelScale / delta)
            else:
                self._dolly_camera(delta)
            self.EndDolly()

    def _mouse_move_event(self, obj: Any, event: Any) -> None:
        interactor = self.GetInteractor()
        event_pos = interactor.GetEventPosition()
        self.FindPokedRenderer(event_pos[0], event_pos[1])

        if self.GetState() == 1:
            # custom rotation
            self._rotate()
        elif self.GetState() == 2:
            self._pan()
        elif self.GetState() == 3:
            # custom spin
            self._spin()
        elif self.GetState() == 4:
            self._dolly()
        else:
            super().OnMouseMove()

        # rescale the pivot sphere upon each mouse move so it is always the same size on screen
        self._scale_pivot_sphere()
        self.last_pos = event_pos
        self.GetInteractor().Render()

    def _normalize_mouse(self, coord: tuple[int, int]) -> None:
        size = self.GetInteractor().GetRenderWindow().GetSize()
        return -1.0 + 2.0 * coord[0] / size[0], -1.0 + 2.0 * coord[1] / size[1]

    def _get_right_v_and_up_v(self, pivot: ArrayLike, camera: pv.Camera) -> tuple[float, float]:
        cam_pos = np.array(camera.GetPosition())
        pivot = np.asarray(pivot)
        vector_to_pivot = pivot - cam_pos

        if self.translation_plane_normal is None:
            view_plane_normal = _normalize(camera.GetViewPlaneNormal())
        else:
            view_plane_normal = _normalize(self.translation_plane_normal)

        l = -np.dot(vector_to_pivot, view_plane_normal)
        view_angle = camera.GetViewAngle() * np.pi / 180

        size = self.GetInteractor().GetRenderWindow().GetSize()
        scalex = size[0] / size[1] * ((2 * l * np.tan(view_angle / 2)) / 2)
        scaley = ((2 * l * np.tan(view_angle / 2)) / 2)

        view_up = camera.GetViewUp()
        right_v = _normalize(np.cross(view_up, view_plane_normal)) * scalex
        up_v = _normalize(np.cross(view_plane_normal, right_v)) * scaley

        return right_v, up_v

    def _translate_camera(self, translation: ArrayLike) -> None:
        camera = self.GetCurrentRenderer().GetActiveCamera()
        cam_pos = np.asarray(camera.GetPosition())
        cam_focal = np.asarray(camera.GetFocalPoint())

        cam_pos_new = cam_pos + translation
        cam_focal_new = cam_focal + translation

        camera.SetPosition(cam_pos_new)
        camera.SetFocalPoint(cam_focal_new)

        if self.GetAutoAdjustCameraClippingRange():
            self.GetCurrentRenderer().ResetCameraClippingRange()

        if self.GetInteractor().GetLightFollowCamera():
            self.GetCurrentRenderer().UpdateLightsGeometryToFollowCamera()

    def _rotate_camera(self, point: ArrayLike, azimuth: float, elevation: float) -> None:
        camera = self.GetCurrentRenderer().GetActiveCamera()
        focal_point = camera.GetFocalPoint()
        view_up = camera.GetViewUp()
        position = camera.GetPosition()

        axis = [0, 0, 0]
        axis[0] = -1 * camera.GetViewTransformMatrix().GetElement(0,0)
        axis[1] = -1 * camera.GetViewTransformMatrix().GetElement(0,1)
        axis[2] = -1 * camera.GetViewTransformMatrix().GetElement(0,2)

        transform = vtkTransform()
        transform.Identity()
        transform.Translate(*point)
        transform.RotateWXYZ(azimuth, view_up)
        transform.RotateWXYZ(elevation, axis)
        transform.Translate(*[-1*x for x in point])

        new_position = transform.TransformPoint(position)
        new_focal_point = transform.TransformPoint(focal_point)

        camera.SetPosition(new_position)
        camera.SetFocalPoint(new_focal_point)

        camera.OrthogonalizeViewUp()

        if self.GetAutoAdjustCameraClippingRange():
            self.GetCurrentRenderer().ResetCameraClippingRange()

        if self.GetInteractor().GetLightFollowCamera():
            self.GetCurrentRenderer().UpdateLightsGeometryToFollowCamera()

    def _pan(self) -> None:
        event_pos = self.GetInteractor().GetEventPosition()
        event_pos_normalized = self._normalize_mouse(event_pos)
        last_pos_normalized = self._normalize_mouse(self.last_pos)

        deltax = event_pos_normalized[0] - last_pos_normalized[0]
        deltay = event_pos_normalized[1] - last_pos_normalized[1]

        camera = self.GetCurrentRenderer().GetActiveCamera()

        right_v, up_v = self._get_right_v_and_up_v(self.pivot, camera)
        offset_v = (-deltax * right_v + (-deltay * up_v))

        self._translate_camera(offset_v)

    def _rotate(self) -> None:
        interactor = self.GetInteractor()
        dx = interactor.GetEventPosition()[0] - interactor.GetLastEventPosition()[0]
        dy = interactor.GetEventPosition()[1] - interactor.GetLastEventPosition()[1]
        size = self.GetCurrentRenderer().GetRenderWindow().GetSize()

        delta_elevation = -20.0 / size[1]
        delta_azimuth = -20.0 / size[0]

        rxf = dx * delta_azimuth * self.GetMotionFactor()
        ryf = dy * delta_elevation * self.GetMotionFactor()

        self._rotate_camera(self.pivot, rxf, ryf)

    def _dolly_camera(self, delta: float) -> None:
        camera = self.GetCurrentRenderer().GetActiveCamera()
        cam_pos = np.array(camera.GetPosition())

        move = np.asarray(self.pivot) - cam_pos
        offset = move * delta * -4
        self._translate_camera(offset)

        if self.GetAutoAdjustCameraClippingRange():
            self.GetCurrentRenderer().ResetCameraClippingRange()

        if self.GetInteractor().GetLightFollowCamera():
            self.GetCurrentRenderer().UpdateLightsGeometryToFollowCamera()

    def _dolly(self) -> None:
        event_pos = self.GetInteractor().GetEventPosition()
        event_pos_normalized = self._normalize_mouse(event_pos)
        last_pos_normalized = self._normalize_mouse(self.last_pos)

        delta = event_pos_normalized[1] - last_pos_normalized[1]
        self._dolly_camera(delta)

    def _spin(self) -> None:
        center = self.GetCurrentRenderer().GetCenter()
        pivot = self.pivot
        event_pos = self.GetInteractor().GetEventPosition()
        last_event_pos = self.last_pos

        new_angle = np.degrees(np.arctan2(event_pos[1] - center[1], event_pos[0]-center[0]))

        old_angle = np.degrees(np.arctan2(last_event_pos[1] - center[1], last_event_pos[0] - center[0]))

        camera = self.GetCurrentRenderer().GetActiveCamera()

        # transform = vtk.vtkTransform()
        # transform.Identity()
        # transform.Translate(*pivot)
        # transform.RotateWXYZ(new_angle - old_angle, camera.GetDirectionOfProjection())
        #
        # position = camera.GetPosition()
        # focal_point = camera.GetFocalPoint()
        # new_position = transform.TransformPoint(position)
        # new_focal_point = transform.TransformPoint(focal_point)
        #
        # camera.SetPosition(new_position)
        # camera.SetFocalPoint(new_focal_point)

        camera.Roll(new_angle - old_angle)
        camera.OrthogonalizeViewUp()

    def remove_observers(self) -> None:
        for obs in self._observers:
            self.RemoveObserver(obs)














import pyvista as pv
import vtk
from pyvista_imgui import ImguiPlotter
from imgui_bundle import immapp, imgui, hello_imgui

sphere = pv.Sphere()


plotter = ImguiPlotter()
plotter.enable_pivot_style()
plotter.add_axes()
plotter.add_mesh(sphere, render=False)

runner_params = hello_imgui.RunnerParams()
runner_params.app_window_params.window_title = "Viewer"
runner_params.app_window_params.window_geometry.size = (1400, 1080)
runner_params.imgui_window_params.show_status_bar = True

def gui():
    hello_imgui.apply_theme(hello_imgui.ImGuiTheme_.imgui_colors_dark)
    vec = imgui.get_main_viewport().pos
    imgui.set_next_window_pos(vec, imgui.Cond_.once)
    imgui.set_next_window_size(imgui.get_main_viewport().size)
    imgui.set_next_window_bg_alpha(1.0)
    imgui.begin("Imgui Plotter", flags=imgui.WindowFlags_.no_bring_to_front_on_focus | imgui.WindowFlags_.no_title_bar | imgui.WindowFlags_.no_decoration | imgui.WindowFlags_.no_resize | imgui.WindowFlags_.no_move)

    # render the plotter's contents here
    plotter.render_imgui()

    imgui.end()
    imgui.show_demo_window()


runner_params.callbacks.show_gui = gui
# IMPORTANT: hello_imgui has a special default window type, that prevets interaction when the Plotter is forced to the background. To prevent this, we disable it here.
runner_params.imgui_window_params.default_imgui_window_type = hello_imgui.DefaultImGuiWindowType.no_default_window
immapp.run(runner_params=runner_params)


from vtkmodules.vtkRenderingOpenGL2 import vtkGenericOpenGLRenderWindow, vtkTextureObject
from vtkmodules.vtkCommonCore import VTK_UNSIGNED_CHAR
import typing as typ


__all__ = ['VTKOpenGLTextureRenderWindow']


class VTKOpenGLTextureRenderWindow(object):
    """
    Class to render the output of one or multiple vtk renderers into an opengl texture object. The texture can be retrieved for use in other visualization packages based on opengl.
    """
    def __init__(self) -> None:
        """
        A specialization of a vtkRenderWindow that renders itself into an opengl texture 
        when calling the 'render' method. 
        The resulting texture an be retrieved for use in external visualization packages based on opengl.

        The size of the resulting texture can be resized dynamically upon calling 'render'
        """
  
        self._tex = None
        self._renderers = []

        self._texture_size = (0, 0)

        self.render_window = vtkGenericOpenGLRenderWindow()
        self.render_window.SetIsCurrent(True)
 
    @property
    def texture_id(self) -> typ.Optional[int]:
        """
        Returns the texture id of the rendererd texture. 
        If nothing has been rendererd yet, None is returned instead.
        """
        if self._tex is None:
            return None # no texture has been created yet
        return self._tex.GetHandle()
    
    @property
    def size(self) -> tuple[int, int]:
        return self.render_window.GetSize()
    
    @size.setter
    def size(self, size: tuple[int, int]) -> None:
        self.render_window.SetSize(int(size[0]), int(size[1]))

    def render(self) -> None:
        """ 
        Renders the vtk output into a texture of appropriate size.
        """
        self.set_viewport_size(self.size)

        vtk_fbo = self.render_window.GetDisplayFramebuffer()
        vtk_fbo.SetContext(self.render_window)
        vtk_fbo.SaveCurrentBindingsAndBuffers()
        vtk_fbo.Bind()
        vtk_fbo.AddColorAttachment(0, self._tex)
        vtk_fbo.AddDepthAttachment()
        vtk_fbo.ActivateBuffer(0)

        self.render_window.Render()
        #self.render_window.WaitForCompletion()

        vtk_fbo.RestorePreviousBindingsAndBuffers()

    def __getattr__(self, attr):
        """
        Makes the object behave like a vtkRenderWindow
        """
        if attr == '__vtk__':
            return lambda t=self.render_window: t
        elif hasattr(self.render_window, attr):
            return getattr(self.render_window, attr)
        else:
            raise AttributeError(self.__class__.__name__ +
                  " has no attribute named " + attr)
        
    def set_viewport_size(self, new_size: tuple[int, int]) -> None:
        """
        Sets the internal texture size in pixels. If the new size is different from
        the current size, a new texture is allocated.

        Parameters
        ----------
        new_size
            the new size of the viewport
        """

        # init the internal render window to use the current (texture) context
        # create a texture object to render into
        if self._tex is None:
            # init the opengl context here
            self.render_window.OpenGLInitContext()
            self.render_window.OpenGLInitState()
            self.render_window.MakeCurrent()

            # setup color texture
            self._tex = vtkTextureObject()
            self._tex.SetContext(self.render_window)

            # allocate the texture object using the initial size
            self._tex.Create2D(new_size[0], new_size[1], 4, VTK_UNSIGNED_CHAR, False)
            self._tex.SetWrapS(vtkTextureObject.ClampToEdge)
            self._tex.SetWrapT(vtkTextureObject.ClampToEdge)
            self._tex.SetMinificationFilter(vtkTextureObject.Linear)
            self._tex.SetLinearMagnification(True)
            self._tex.Bind()
            self._tex.SendParameters()

        current_size = self.size
        # if nothing has changed just return
        if current_size == new_size or any(s <= 0 for s in new_size):
            return
        
        self.size = new_size
        self._tex.Resize(new_size[0], new_size[1])



import pyvista as pv
from pyvista_imgui import ImguiPlotter

sphere = pv.Sphere()

plotter = ImguiPlotter(imgui_backend='imgui_bundle')
plotter.add_axes()
plotter.add_mesh(sphere, render=False)
plotter.show()


import typing as typ
from .imgui_render_window import RendererBackend, register_backend
from vtkmodules.vtkCommonCore import vtkCommand
from vtkmodules.vtkRenderingCore import vtkRenderWindow, vtkRenderWindowInteractor
import typing as typ

try:
    from imgui_bundle import imgui
except ImportError:
    imgui = None

if imgui: 
    _Key = imgui.Key
    _keysyms = {
        _Key.backspace: ('BackSpace', None),
        _Key.tab: ('Tab', None),
        _Key.enter: ('Return', None),
        _Key.keypad_enter: ('Return', None),
        _Key.left_shift: ('Shift_L', None),
        _Key.left_ctrl: ('Control_L', None),
        _Key.left_alt: ('Alt_L', None),
        _Key.pause: ('Pause', None),
        _Key.caps_lock: ('Caps_Lock', None),
        _Key.escape: ('Escape', None),
        _Key.space: ('space', ' '),
        # Key.Key_Prior : 'Prior',
        # Key.Key_Next : 'Next',
        _Key.end: ('End', None),
        _Key.home: ('Home', None),
        _Key.left_arrow: ('Left', None),
        _Key.up_arrow: ('Up', None),
        _Key.right_arrow: ('Right', None),
        _Key.down_arrow: ('Down', None),
        _Key.insert: ('Insert', None),
        _Key.delete: ('Delete', None),
        _Key._0: ('0', '0'),
        _Key._1: ('1', '1'),
        _Key._2: ('2', '2'),
        _Key._3: ('3', '3'),
        _Key._4: ('4', '4'),
        _Key._5: ('5', '5'),
        _Key._6: ('6', '6'),
        _Key._7: ('7', '7'),
        _Key._8: ('8', '8'),
        _Key._9: ('9', '9'),
        _Key.a: ('a', 'a'),
        _Key.b: ('b', 'b'),
        _Key.c: ('c', 'c'),
        _Key.d: ('d', 'd'),
        _Key.e: ('e', 'e'),
        _Key.f: ('f', 'f'),
        _Key.g: ('g', 'g'),
        _Key.h: ('h', 'h'),
        _Key.i: ('i', 'i'),
        _Key.j: ('j', 'j'),
        _Key.k: ('k', 'k'),
        _Key.l: ('l', 'l'),
        _Key.m: ('m', 'm'),
        _Key.n: ('n', 'n'),
        _Key.o: ('o', 'o'),
        _Key.p: ('p', 'p'),
        _Key.q: ('q', 'q'),
        _Key.r: ('r', 'r'),
        _Key.s: ('s', 's'),
        _Key.t: ('t', 't'),
        _Key.u: ('u', 'u'),
        _Key.v: ('v', 'v'),
        _Key.w: ('w', 'w'),
        _Key.x: ('x', 'x'),
        _Key.y: ('y', 'y'),
        _Key.z: ('z', 'z'),
        _Key.keypad_add: ('plus', '+'),
        _Key.right_bracket: ('plus', '+'),
        _Key.minus: ('minus', '-'),
        _Key.period: ('period', '~'),
        _Key.slash: ('slash', '/'),
        _Key.f1: ('F1', None),
        _Key.f2: ('F2', None),
        _Key.f3: ('F3', None),
        _Key.f4: ('F4', None),
        _Key.f5: ('F5', None),
        _Key.f6: ('F6', None),
        _Key.f7: ('F7', None),
        _Key.f8: ('F8', None),
        _Key.f9: ('F9', None),
        _Key.f10: ('F10', None),
        _Key.f11: ('F11', None),
        _Key.f12: ('F12', None),
        _Key.num_lock: ('Num_Lock', None),
        _Key.scroll_lock: ('Scroll_Lock', None),
    }

    _ignore_keys = {
        _Key.mouse_left,
        _Key.mouse_right,
        _Key.mouse_middle,
        _Key.mouse_wheel_x,
        _Key.mouse_wheel_y,
    }


__all__ = ['RendererBackendImguiBundle']

@register_backend("imgui_bundle")
class RendererBackendImguiBundle(RendererBackend):
    def __init__(self, 
                 interactor: vtkRenderWindowInteractor, 
                 render_window: vtkRenderWindow,
                 border = False) -> None:
        if not imgui:
            raise ModuleNotFoundError(f"{self.__class__.__name__} requires the 'imgui_bundle' package.")
        super().__init__(interactor, render_window, border=border)

    def render(self, size: typ.Optional[tuple[int, int]] = None):
        # get the maximum available size
        size = size or imgui.get_content_region_avail()
        size = (size.x, size.y)
        self.render_window.size = size
        self.render_window.render()

        # adjust the size of this interactor as well
        self.interactor.SetSize(int(size[0]), int(size[1]))
        self.interactor.ConfigureEvent()

        # render the texture with the vtk output into an image
        imgui.push_style_var(imgui.StyleVar_.window_padding, (0, 0))
        # make the image unscrollable to ensure correct mouse behavior
        no_scroll_flags = imgui.WindowFlags_.no_scrollbar | imgui.WindowFlags_.no_scroll_with_mouse
        imgui.begin_child("##Viewport", size, self.border, no_scroll_flags)
        imgui.image(self.render_window.texture_id, 
                    imgui.get_content_region_avail(), 
                    (0, 1), (1, 0))
        # process the events of this widget
        self.process_events()
        imgui.end_child()
        imgui.pop_style_var()

    def _process_mouse_events(self, io):
        if imgui.is_window_hovered():
            if io.mouse_clicked[imgui.MouseButton_.left]:
                self.interactor.InvokeEvent(vtkCommand.LeftButtonPressEvent)
            elif io.mouse_clicked[imgui.MouseButton_.right]:
                self.interactor.InvokeEvent(vtkCommand.RightButtonPressEvent)
            elif io.mouse_clicked[imgui.MouseButton_.middle]:
                self.interactor.InvokeEvent(vtkCommand.MiddleButtonPressEvent)
            elif io.mouse_wheel > 0:
                self.interactor.InvokeEvent(vtkCommand.MouseWheelForwardEvent)
            elif io.mouse_wheel < 0:
                self.interactor.InvokeEvent(vtkCommand.MouseWheelBackwardEvent)

        if io.mouse_released[imgui.MouseButton_.left]:
            self.interactor.InvokeEvent(vtkCommand.LeftButtonReleaseEvent)
        elif io.mouse_released[imgui.MouseButton_.right]:
            self.interactor.InvokeEvent(vtkCommand.RightButtonReleaseEvent)
        elif io.mouse_released[imgui.MouseButton_.middle]:
            self.interactor.InvokeEvent(vtkCommand.MiddleButtonReleaseEvent)
        
        self.interactor.InvokeEvent(vtkCommand.MouseMoveEvent)

    def _process_keyboard_events(self, xpos, ypos, ctrl, shift):
        if imgui.is_window_hovered():
            key_start = imgui.Key.named_key_begin.value
            key_end = imgui.Key.named_key_end.value
            
            # for each valid key check if it's state changed and emit the appropriate vtk event
            for k in range(key_start, key_end):
                k = imgui.Key(k)
                if k in _ignore_keys:
                    continue

                if imgui.is_key_pressed(k) or imgui.is_key_released(k):
                    try:
                        keysym, keychar = _keysyms[k]
                    except (KeyError):
                        keysym = None
                        keychar = None

                    keychar = keychar or '\0'

                    self.interactor.SetEventInformationFlipY(xpos, ypos, ctrl, shift, keychar, 0, keysym)
                    if imgui.is_key_pressed(k):
                        self.interactor.KeyPressEvent()
                        self.interactor.CharEvent()
                    else:
                        self.interactor.KeyReleaseEvent()

    def process_events(self):
        """
        Handle events by passing them to the underlying vtk interactor. This method is called automatically
        on rendering.
        """
        # do nothing as long as the mouse pointer is not within the current window or it is not focussed
        if not imgui.is_window_focused() and not imgui.is_window_hovered():
            return
        io = imgui.get_io()
        io.config_windows_move_from_title_bar_only = True # do not drag the window when clicking on the image
        viewport_pos = imgui.get_cursor_start_pos()

        xpos = int(io.mouse_pos.x - viewport_pos.x)
        ypos = int(io.mouse_pos.y - viewport_pos.y)

        ctrl = io.key_ctrl
        shift = io.key_shift
        
        repeat = 0
        if io.mouse_double_clicked[0] or io.mouse_double_clicked[1] or io.mouse_double_clicked[2]:
            repeat = 1

        if xpos < 0 or ypos < 0:
            return 
        
        self.interactor.SetEventInformationFlipY(xpos, ypos, ctrl, shift, chr(0), repeat, None)

        self._process_mouse_events(io)
        self._process_keyboard_events(xpos, ypos, ctrl, shift)

    def show(self, 
             title: typ.Optional[str] = None, 
             window_size: tuple[int, int] = (1400, 1080)):
        """
        show method for the imgui-bundle package.

        Parameters
        ----------
        title, optional
            The title of the window, if None, a default title is used.
        window_size, optional
            The size of the displayed window, by default (1400, 1080)
        """

        from imgui_bundle import immapp, hello_imgui, imgui     
        runner_params = hello_imgui.RunnerParams()
        runner_params.app_window_params.window_title = title or "ImguiPlotter"
        runner_params.app_window_params.window_geometry.size = window_size
        runner_params.imgui_window_params.show_status_bar = True

        def gui():
            hello_imgui.apply_theme(hello_imgui.ImGuiTheme_.imgui_colors_dark)
            vec = imgui.get_main_viewport().pos
            imgui.set_next_window_pos(vec, imgui.Cond_.once)
            imgui.set_next_window_size(imgui.get_main_viewport().size)
            imgui.set_next_window_bg_alpha(1.0)
            imgui.begin("Vtk Viewer", flags=imgui.WindowFlags_.no_bring_to_front_on_focus | imgui.WindowFlags_.no_title_bar | imgui.WindowFlags_.no_decoration | imgui.WindowFlags_.no_resize | imgui.WindowFlags_.no_move)
            self.render()
            imgui.end()

        runner_params.callbacks.show_gui = gui
        runner_params.imgui_window_params.default_imgui_window_type = hello_imgui.DefaultImGuiWindowType.no_default_window
        immapp.run(runner_params=runner_params)
        
      


import pyvista as pv
from pyvista_imgui import ImguiPlotter
from threading import Thread
import time

sphere = pv.Sphere()

plotter = ImguiPlotter()
plotter.add_axes()
plotter.add_mesh(sphere, render=False)

done = False

def plot():
    try:
        plotter.show()
    finally:
        done = True

t = Thread(target=plot)
t.start()


while not done:
    time.sleep(0.5)
    print("sleeping")
t.join()




from .texture_render_window import VTKOpenGLTextureRenderWindow
from vtkmodules.vtkRenderingUI import vtkGenericRenderWindowInteractor
from vtkmodules.vtkCommonCore import vtkCommand
from vtkmodules.vtkRenderingCore import vtkRenderWindow, vtkRenderWindowInteractor

import enum
from abc import ABC, abstractmethod
from weakref import WeakValueDictionary

import typing as typ

__all__ = ['VTKImguiRenderWindowInteractor', 'RendererBackend', 'register_backend']


class RendererBackend(object):
    _backends = WeakValueDictionary()
    def __init__(self, 
                 interactor: vtkRenderWindowInteractor, 
                 render_window: vtkRenderWindow,
                 border = False) -> None:
        self.interactor = interactor
        self.render_window = render_window
        self.border = border

    @abstractmethod
    def render(self) -> None:
        pass

    @abstractmethod
    def process_events(self):
        pass

    @abstractmethod
    def show(title: typ.Optional[str] = None, 
             window_size = None):
        pass

    @classmethod
    def _register(cls, clss, name):
        cls._backends[name] = clss

    @classmethod
    def from_name(cls, name, interactor, render_window, border=False):
        return cls._backends[name](interactor, render_window, border=border)
        

def register_backend(name) -> typ.Callable:
    def _backend_decoractor(backend_cls: typ.Type[RendererBackend]) -> typ.Type[RendererBackend]:
        assert issubclass(backend_cls, RendererBackend)
        RendererBackend._register(backend_cls, name)
        return backend_cls
    return _backend_decoractor


class VTKImguiRenderWindowInteractor(object):
    """
    A RenderWindowInteractor to integrate VTK's rendering within an existing imgui ui.
    It uses the VTKOpenGLTextureRenderWindow to render VTK's output into an openGl-texture,
    which is displayed as an image within imgui.

    All interaction events are passed to the underlying interactor to be handled, so all
    mouse interactions are the same as in regular VTK.

    Parameters
    ----------
    border, optional
        display a border around the created imgui widget, False by default.
    """
    def __init__(self, 
                 border: bool = False,
                 imgui_backend = None) -> None:
        if imgui_backend is None:
            imgui_backend = 'imgui_bundle'

        self.ren_win = VTKOpenGLTextureRenderWindow()

        self.interactor = vtkGenericRenderWindowInteractor()

        self.imgui_backend = RendererBackend.from_name(imgui_backend, self.interactor, self.ren_win, border=border)

        # do not render unless explicitly requested, as imgui has control over the event loop
        self.interactor.EnableRenderOff()
        self.ren_win.SetInteractor(self.interactor)
    
    def __getattr__(self, attr):
        """
        Makes this object behave like a vtkRenderWindowInteractor
        """
        if attr == '__vtk__':
            return lambda t=self.interactor: t
        elif hasattr(self.interactor, attr):
            return getattr(self.interactor, attr)
        else:
            raise AttributeError(self.__class__.__name__ +
                  " has no attribute named " + attr)

    def render_imgui(self, size: typ.Optional[tuple[int, int]] = None) -> None:
        """
        Renders the contents of the internal render window into an existing imgui ui 
        with the specified size. This method should be called when building 
        and updating the ui.

        Parameters
        ----------
        size, optional
            the size of the result in the ui in pixels, 
            if None (default) the maximum available size is used.
        """
        self.imgui_backend.render(size)

    def process_events(self) -> None:
        """
        Handle events by passing them to the underlying vtk interactor. This method is called automatically
        on rendering.
        """
        self.imgui_backend.process_events()




try:
    from importlib.metadata import version

    __version__ = version("pyvista_imgui")
except Exception:  # pragma: no cover # pylint: disable=broad-exception-caught
    try:
        from ._version import __version__
    except ImportError:
        __version__ = '0.0.0'

from .texture_render_window import *
from .imgui_render_window import *
from .plotting import *
from ._backend_imgui_bundle import *
from ._backend_pyimgui import *


from __future__ import annotations
import pyvista as pv
from .imgui_render_window import VTKImguiRenderWindowInteractor
from pyvista import global_theme
from pyvista.plotting.render_window_interactor import RenderWindowInteractor
import typing as typ
from functools import wraps
import contextlib
import numpy as np
from collections.abc import Sequence
from vtkmodules.vtkCommonDataModel import vtkPolyData
from vtkmodules.vtkRenderingCore import vtkGlyph3DMapper

__all__ = ['ImguiPlotter']

# monkey patch the pointgaussianmapper to disable the automatic rgba conversion of the colors
from pyvista.plotting.mapper import PointGaussianMapper

def _as_rgba_disabled(self):
    """Patched method to disable the rgba conversion for the PointGaussianMapper
    """
    if self.color_mode == 'direct':
        return

PointGaussianMapper.as_rgba = _as_rgba_disabled


@contextlib.contextmanager
def _no_base_plotter_init():
    init = pv.BasePlotter.__init__
    pv.BasePlotter.__init__ = lambda *args, **kwargs: None
    try:
        yield
    finally:
        pv.BasePlotter.__init__ = init

class ImguiPlotter(VTKImguiRenderWindowInteractor, pv.BasePlotter):
    """
    This class extends pyvista's BasePlotter making it available
    as a DearImGui widget. 

    To display this plotter withing an existing imgui application, just call
    the 'render'-method when building the ui. 

    Alternatively, the plotter also supports rendering as a standalone 
    imgui-application by calling the 'show'-method. It will display the plotter
    as a widget inside a minimal window-application.

    Parameters
    ----------
    multi_samples, optional
        The number of multi-samples used to mitigate aliasing. 4 is a
        good default but 8 will have better results with a potential
        impact on performance.. Defaults to None.
    line_smoothing, optional
        Enable line smoothing, by default False
    point_smoothing, optional
        Enable point smoothing, by default False
    polygon_smoothing, optional
        Enable polygon smoothing, by default False
    background, optional
        Spawn the viewer in a background thread to allow the main python interpreter to continue. 
        Only works when calling 'show'
    imgui_backend, optional
        The imgui backend to use for the ui. Use either 'pyimgui' or 'imgui_bundle'.
    """

    def __init__(self, 
                 title=None,
                 window_size=None,
                 line_smoothing: bool = False,
                 point_smoothing: bool = False,
                 polygon_smoothing: bool = False,
                 background: bool = False,
                 imgui_backend: typ.Optional[str] = None,
                 **kwargs):
        
        self._initialized = False
        
        # imgui has it's own border functionality so ignore the vtk one
        border = kwargs.pop("border", False)
        with _no_base_plotter_init():
            VTKImguiRenderWindowInteractor.__init__(self, border=border, imgui_backend=imgui_backend)
        pv.BasePlotter.__init__(self, **kwargs)
        self.title = title or "ImguiPlotter"
        self.suppress_rendering = True # disable rendering as the event loop is controlled by imgui

        self.render_window.SetMultiSamples(0)
        if line_smoothing:
            self.render_window.LineSmoothingOn()
        if point_smoothing:
            self.render_window.PointSmoothingOn()
        if polygon_smoothing:
            self.render_window.PolygonSmoothingOn()

        for renderer in self.renderers:
            self.render_window.AddRenderer(renderer)

        # Add the shadow renderer to allow us to capture interactions within
        # a given viewport
        # https://vtk.org/pipermail/vtkusers/2018-June/102030.html
        number_or_layers = self.render_window.GetNumberOfLayers()
        current_layer = self.renderer.GetLayer()
        self.render_window.SetNumberOfLayers(number_or_layers + 1)
        self.render_window.AddRenderer(self.renderers.shadow_renderer)
        self.renderers.shadow_renderer.SetLayer(current_layer + 1)
        self.renderers.shadow_renderer.SetInteractive(False)  # never needs to capture
        
        self.background_color = self.theme.background
        self._setup_interactor()

        # # Add the shadow renderer to allow us to capture interactions within
        # # a given viewport
        # # https://vtk.org/pipermail/vtkusers/2018-June/102030.html
        number_or_layers = self.render_window.GetNumberOfLayers()
        current_layer = self.renderer.GetLayer()
        self.render_window.SetNumberOfLayers(number_or_layers + 1)
        self.render_window.AddRenderer(self.renderers.shadow_renderer)
        self.renderers.shadow_renderer.SetLayer(current_layer + 1)
        self.renderers.shadow_renderer.SetInteractive(False)  # never needs to capture


        # Set window size
        self._window_size_unset = False
        if window_size is None:
            self.window_size = self._theme.window_size
            if self.window_size == pv.plotting.themes.Theme().window_size:
                self._window_size_unset = True
        else:
            self.window_size = window_size

        # # Set camera widget based on theme. This requires that an
        # # interactor be present.
        if self.theme._enable_camera_orientation_widget:
            self.add_camera_orientation_widget()

        # Set background
        self.set_background(self._theme.background)

        if self._theme.depth_peeling.enabled:
            if self.enable_depth_peeling():
                for renderer in self.renderers:
                    renderer.enable_depth_peeling()

        # set anti_aliasing based on theme
        if self.theme.anti_aliasing:
            self.enable_anti_aliasing(self.theme.anti_aliasing)

        self._run_background = background
        self._thread = None

        # self._app = None
        # Set some private attributes that let BasePlotter know
        # that this is safely rendering
        self._first_time = False  # Crucial!
        self._initialized = True

    def _setup_interactor(self) -> None:
        self.iren = RenderWindowInteractor(
            self, interactor=self.render_window.GetInteractor()
        )
        self.iren.interactor.RemoveObservers(
            "MouseMoveEvent"
        )  # slows window update?
        self.iren.initialize()
        self.enable_trackball_style()
        self.iren.add_observer("KeyPressEvent", self.key_press_event)


    def __del__(self):
        # We have to check here if the plotter was only partially initialized
        self.deep_clean()
        if self._initialized:
            del self.renderers

    @property
    def window_size(self) -> tuple[int, int]:  # numpydoc ignore=RT01
        """Return the render window size in ``(width, height)``.

        Examples
        --------
        Change the window size from ``200 x 200`` to ``400 x 400``.

        >>> import pyvista as pv
        >>> pl = pv.Plotter(window_size=[200, 200])
        >>> pl.window_size
        [200, 200]
        >>> pl.window_size = [400, 400]
        >>> pl.window_size
        [400, 400]

        """
        return tuple(self.render_window.GetSize())

    @window_size.setter
    def window_size(self, window_size):  # numpydoc ignore=GL08
        self.render_window.SetSize(window_size[0], window_size[1])
        self._window_size_unset = False
        #self.render() # rendering is controlled by imgui so do not do that here!

    def close(self, render=False):
        """Override the BasePlotter's close method to ensure correct behavior.

        Parameters
        ----------
        render : bool
            Unused argument.

        """
        from pyvista.plotting.widgets import WidgetHelper
        from pyvista.plotting.plotting import _ALL_PLOTTERS
        
        # optionally run just prior to exiting the plotter
        if self._before_close_callback is not None:
            self._before_close_callback(self)
            self._before_close_callback = None

        # must close out widgets first
        WidgetHelper.close(self)
        # Renderer has an axes widget, so close it
        self.renderers.close()
        self.renderers.remove_all_lights()

        # reset scalar bars
        self.scalar_bars.clear()
        self.mesh = None
        self.mapper = None

        if hasattr(self, 'textActor'):
            del self.textActor

        # end movie
        if hasattr(self, 'mwriter'):
            try:
                self.mwriter.close()
            except BaseException:
                pass

        # Remove the global reference to this plotter unless building the
        # gallery to allow it to collect.
        if not pv.BUILDING_GALLERY:
            if _ALL_PLOTTERS is not None:
                _ALL_PLOTTERS.pop(self._id_name, None)

        # this helps managing closed plotters
        self._closed = True 

    @wraps(pv.BasePlotter.add_actor)
    def add_actor(self, *args, **kwargs):
        """
        Override to ignore the 'render' argument
        """
        kwargs["render"] = False # never render as this is not controlled by vtk any more
        return pv.BasePlotter.add_actor(self, *args, **kwargs)
    
    @wraps(pv.BasePlotter.set_background)
    def set_background(self, color, top=None, right=None, side=None, corner=None, all_renderers=True):
        # the background color is not set unless the top color is set as well to enforce it 
        super().set_background(color, top=top or color, right=right, side=side, corner=corner, all_renderers=all_renderers)

    def show(self, 
             window_size: tuple[int, int] = None,
             before_close_callback: typ.Optional[typ.Callable] = None,
            **kwargs: typ.Any):
        """
        Show this plotter as a standalone imgui application. This call blocks until
        the resulting window is closed again.
        To integrate within an existing imgui context use render() within 
        the rendering loop.

        Parameters
        ----------
        window_size, optional
            The window size. Defaults to (1400, 1080).
        before_close_callback, optional
            An optional callback to execute before closing, by default None.
        kwargs: Exists to ensure compatiblity with the BasePlotter interface. 
                Any additional keywords are ignored for this plotter.
        """
        if before_close_callback is None:
            before_close_callback = global_theme._before_close_callback
        self._before_close_callback = before_close_callback

        if window_size is None:
            window_size = self.window_size
        else:
            self._window_size_unset = False
        self.render_window.SetSize(window_size[0], window_size[1])

        def _show():
            self.imgui_backend.show(title=self.title, window_size=window_size)

        if self._run_background:
            from threading import Thread
            self._thread = Thread(target=_show)
            self._thread.start()
        else:
            _show()
    
    def enable_pivot_style(self, pivot_sphere_radius: float = 0.25, pivot_sphere_resolution: int = 16) -> None:
        from .interactor_style_pivot import InteractorStylePivot
        pivot = InteractorStylePivot(self, pivot_sphere_radius=pivot_sphere_radius, pivot_sphere_resolution=pivot_sphere_resolution)
        return self._update_style('Pivot', pivot) 
    
    def _update_style(self, name, style_class):
        self.iren._style_class = style_class
        self.iren._style = name
        ret = self.iren.update_style()
        if getattr(self, '_fixed_clipping_range', False):
            self._style.SetAutoAdjustCameraClippingRange(True)
        return ret

    def add_glyph_mapper(self, 
                        dataset: pv.DataSet, 
                        orient: bool | str = True, 
                        scale: bool | str | typ.ArrayLike = True, 
                        factor: float = 1.0, 
                        geom: typ.Optional[pv.Dataset | tuple[pv.DataSet, ...]] = None,
                        indices: typ.Optional[tuple[float, ...]] = None, 
                        tolerance: typ.Optional[float] = None, 
                        absolute: bool = False, 
                        clamping: bool = False,
                        rng: typ.Optional[tuple[float, float]] = None, 
                        reset_camera: bool = False, 
                        name: typ.Optional[str] = None, 
                        pickable: bool = True) -> pv.Actor:
        """
        Add a Mapper copying a geometric representation (called a glyph) to the input dataset.
        This is similar to using the mesh.glyph filter in pyvista, except that all the glyph creation is 
        done at rendering time of the gpu resulting in saved memory and speedup.

        The glyphs may be oriented along the input vectors, and scaled according to scalar data or vector
        magnitude. Passing a table of glyphs to choose from based on
        scalars or vector magnitudes is also supported.  The arrays
        used for ``orient`` and ``scale`` must be either both point data
        or both cell data.

        Args:
            dataset: The dataset to build the glyphs from
            orient: If ``True``, use the active vectors array to orient the glyphs.
                    If string, the vector array to use to orient the glyphs. Defaults to True.
            scale: If ``True``, use the active scalars to scale the glyphs.
                   If string, the scalar array to use to scale the glyphs. Defaults to True.
            factor: Scale factor applied to scaling array. Defaults to 1.0.
            geom: The geometry to use for the glyph. If missing, an arrow glyph
                  is used. If a sequence is passed, the datasets inside define a table of
                  geometries to choose from based on scalars or vectors. In this
                  case a sequence of numbers of the same length must be passed as
                  ``indices``. The values of the range (see ``rng``) affect lookup
                  in the table. Defaults to None.
            indices: Specifies the index of each glyph in the table for lookup in case
                     ``geom`` is a sequence. If given, must be the same length as
                     ``geom``. If missing, a default value of ``range(len(geom))`` is
                     used. Indices are interpreted in terms of the scalar range
                     (see ``rng``). Ignored if ``geom`` has length 1.. Defaults to None.
            tolerance: Specify tolerance in terms of fraction of bounding box length.
                       Float value is between 0 and 1. Default is None. If ``absolute``
                       is ``True`` then the tolerance can be an absolute distance.
                       If ``None``, points merging as a preprocessing step is disabled. Defaults to None.
            absolute: Control if ``tolerance`` is an absolute distance or a fraction. Defaults to False.
            clamping: Turn on/off clamping of "scalar" values to range. Defaults to False.
            rng: Set the range of values to be considered by the filter when scalars values are provided. Defaults to None.
            reset_camera: Reset the camera on insertion of the actor. Defaults to False.
            name: Name of the actor. Defaults to None.
            pickable: Set whether the created actor is pickable. Defaults to True.

        Returns:
            Actor representing the glphed geometry
        """

        # Clean the points before glyphing
        if tolerance is not None:
            small = pv.PolyData(dataset.points)
            small.point_data.update(dataset.point_data)
            dataset = small.clean(point_merging=True, merge_tol=tolerance,
                                  lines_to_points=False, polys_to_lines=False,
                                  strips_to_polys=False, inplace=False,
                                  absolute=absolute)

        # Make glyphing geometry if necessary
        if geom is None:
            geom = pv.Arrow()

        # Check if a table of geometries was passed
        if isinstance(geom, (np.ndarray, Sequence)):
            if indices is None:
                # use default "categorical" indices
                indices = np.arange(len(geom))
            if not isinstance(indices, (np.ndarray, Sequence)):
                raise TypeError('If "geom" is a sequence then "indices" must '
                                'also be a sequence of the same length.')
            if len(indices) != len(geom) and len(geom) != 1:
                raise ValueError('The sequence "indices" must be the same length '
                                 'as "geom".')
        else:
            geom = [geom]
        if any(not isinstance(subgeom, vtkPolyData) for subgeom in geom):
            raise TypeError('Only PolyData objects can be used as glyphs.')

        # Prepare the mapper
        mapper = vtkGlyph3DMapper()
        if len(geom) == 1:
            # use a single glyph, ignore indices
            mapper.SetSourceData(geom[0])
        else:
            for index, subgeom in zip(indices, geom):
                mapper.SetSourceData(index, subgeom)
            mapper.SetSourceIndexing(dataset.active_scalars is not None)

        scale_array_name = None
        if isinstance(scale, str):
            mapper.SetScaleArray(scale)
            scale_array_name = scale
            #dataset.set_active_scalars(scale, 'cell')
            scale = True
        if scale:
            if scale_array_name is not None:
                scale_array = dataset.get_array(scale_array_name, preference='cell')

                if scale_array.ndim > 1:
                    mapper.SetScaleModeToScaleByVectorComponents()
                else:
                    mapper.SetScaleModeToScaleByMagnitude()
        else:
            mapper.SetScaleModeToNoDataScaling()

        if isinstance(orient, str):
            dataset.active_vectors_name = orient
            orient = True

        if scale and orient:
            if (dataset.active_vectors_info.association == pv.FieldAssociation.CELL
                    and dataset.active_scalars_info.association == pv.FieldAssociation.CELL
            ):
                source_data = dataset.cell_centers()
            elif (dataset.active_vectors_info.association == pv.FieldAssociation.POINT
                  and dataset.active_scalars_info.association == pv.FieldAssociation.POINT
            ):
                source_data = dataset
            else:
                raise ValueError("Both ``scale`` and ``orient`` must use "
                                 "point data or cell data.")
        else:
            source_data = dataset

        if rng is not None:
            mapper.SetRange(rng)
        mapper.SetOrient(orient)
        mapper.SetInputData(source_data)
        mapper.ScalingOn()
        #alg.SetVectorModeToUseVector()
        mapper.SetScaleFactor(factor)
        mapper.SetClamping(clamping)

        if name is None:
            name = '{}({})'.format(type(dataset).__name__, dataset.memory_address)

        actor, _ = self.add_actor(mapper,
                                  reset_camera=reset_camera,
                                  name=name,
                                  pickable=pickable, render=False)
        return actor
    
    def add_pointcloud(self,
                       point_cloud: pv.PointSet | pv.PolyData,
                       scalars: typ.Optional[str | typ.ArrayLike] = None,
                       **kwargs: typ.Any) -> pv.Actor:
        """
        Add a dataset (PointSet, PolyData) of any array-like object representing a point cloud  to the
        viewer. This is a specialized version of the add_points function used specifically for point clouds.
        The dataset does not require cells to be defined.

        Args:
            point_cloud: The point cloud to add.
            scalars: The scalars to use for coloring or an array of scalar values.
            **kwargs: Additional keyword arguments are passed to the ImguiPlotter.add_points function

        Returns:
            an Actor instances representing the point cloud in the viewer.
        """

        if not pv.is_pyvista_dataset(point_cloud):
            # custom conversion
            if hasattr(point_cloud, "to_vtk"):
                point_cloud = point_cloud.to_vtk(point_cloud, scalars=scalars, add_cells=False)
    
        # remove some kwargs from the arguments as they cause strange
        # behavior for point clouds or make things very sluggish:
        # -----------------------------------------------------------------------------------------------------------------
            
        # the actual size of the points is controlled via the point size attribute of prop as the option to add points
        # uses this to scale the point gaussians (slow!)
        point_size = kwargs.pop('point_size', 0.0)

        # the default is to render the points as splats so explicitly set this to False
        render_points_as_spheres = kwargs.pop('render_points_as_spheres', False)

        opacity = kwargs.pop('opacity', 1.0)

        # interpolate_before_map causes the plot function to segfault as vtk does not support this so remove it
        kwargs.pop('interpolate_before_map', None)

        # pass the arguments to the regular add_points function
        actor = self.add_points(point_cloud,
                                scalars=scalars,
                                style='points_gaussian',
                                point_size=0,
                                render_points_as_spheres=render_points_as_spheres,
                                interpolate_before_map=False,
                                opacity=1.0,
                                **kwargs)
        
        # set the point size directly in the prop
        actor.prop.point_size = point_size
        # add_points strangely sets the opacity to 0.9999 for some reason. This causes the depth buffer to break, so
        # we revert this setting
        actor.prop.opacity = opacity

        return actor